'use strict';

const dashboardVersion = '0.1.2a';

let serverAddress;
let context;
let lastHeardLimit;
let dstarnowLimit;
let logLevel;

let logger;

let noraapi;
let kdkapi;
let currentSelectedFunction = 'home';
let dashboardInfo;

$(document).ready(()=>{
    initialize();
});

function initialize() {
    Utils.injectStringIncludes();

    if(!logLevelConfig || "" === logLevelConfig)
        logLevel = 'info';
    else
        logLevel = logLevelConfig;

    log.setDefaultLevel(logLevel);
    logger = log.getLogger(Utils.getFileName());
    logger.setLevel(logLevel);

    let serverPort = 0;
    if(!serverPortConfig || serverPortConfig <= 0)
        serverPort = 3000;
    else
        serverPort = serverPortConfig;

    if(serverAddressConfig && "" !== serverAddressConfig)
        serverAddress = 'http://' + serverAddressConfig + ':' + serverPort;
    else if(location.protocol === 'file:')
        serverAddress = 'http://localhost:' + serverPort;
    else
        serverAddress = location.host;

    if(!contextConfig || "" === contextConfig)
        context = "/nora";
    else
        context = contextConfig;

    if(!lastHeardLimitConfig)
        lastHeardLimit = 20;
    else if(lastHeardLimitConfig <= 0)
        lastHeardLimit = 1;
    else
        lastHeardLimit = lastHeardLimitConfig;

    if(!dstarnowLimitConfig)
        dstarnowLimit = 15;
    else if(dstarnowLimitConfig <= 0)
        dstarnowLimit = 1;
    else
        dstarnowLimit = dstarnowLimitConfig;

    const navbarText = $('#main_navbar_text');
    if(navbarText.exists())
        navbarText.html('dashboard v' + dashboardVersion + '<br />Server -> ' + serverAddress + context);

    setupNoraAPI(serverAddress, context);

    setupKdkAPI('http://k-dk.net:8080', '/kdkapi');
}

function showTopMessage(type, message) {
    let alertType;
    let messageHeader;
    switch(type){
        case "success":
            alertType = "alert alert-success";
            messageHeader = "Well done!";
            break;
        case "info":
            alertType = "alert alert-info";
            messageHeader = "Okey!";
            break;
        case "warn":
            alertType = "alert alert-warning";
            messageHeader = "Warning!";
            break;
        case "danger":
        case "error":case "err":
            alertType = "alert alert-danger";
            messageHeader = "Error!";
            break;
        default:
            alertType = "alert alert-info";
            messageHeader = "";
            break;
    }

    clearTopMessage();

    $('body').prepend(
        $('<div/>',{
            "class":alertType,
            "role":"alert"
        }).html("<strong>" + messageHeader + "</strong><p>" + message + "</p>")
    );
}

function clearTopMessage() {
    $('body').children().remove('[class*=alert]');
}

function setupKdkAPI(serverAddress, context) {
    kdkapi = new KdkAPI(
        serverAddress, context, 'error'
    );

    kdkapi.connect(
        () => {
            kdkapi.onUpdateHeardEntryEvent = onJDSTARNowReceiveUpdateHeardEntry;
            kdkapi.onUpdateHeardEntriesEvent = onJDSTARNowReceiveUpdateHeardEntries;
            setTimeout(() =>{
                kdkapi.enableJARLDSTARNow = true;
                kdkapi.requestHeardLog();
            }, 1000);
        },
        () => {
        },
        () => {
            showTopMessage("error", "KdkAPI Connection timeout");
        },
        (err) => {
            showTopMessage("error", "KdkAPI Connection error = " + err);
        }
    );
}

function onJDSTARNowReceiveUpdateHeardEntries(data) {
    $(data).each((index, entry) =>{
        onJDSTARNowReceiveUpdateHeardEntry(entry);
    });
}

function onJDSTARNowReceiveUpdateHeardEntry(data) {
    logger.trace(data);

    const heardlog = $("#jdstarnow_heard_entries");
    if(!heardlog.exists()){
        logger.error("HeardLog not found.");
        return;
    }

    let entry =
        $(heardlog).find("td[name='myLong']")
        .find("a")
        .filter(function() {
            return $(this).text() === data.myCallsign;
        }).parent();
    if(entry.exists()){
        entry = entry.parent();
        
        entry.find("[name='time']").text(
            moment.unix(data.time / 1000).format("MM/DD HH:mm:ss")
        );
        entry.find("[name='rpt1']").text(data.repeater1Callsign);
        entry.find("[name='rpt2']").text(data.repeater2Callsign);
        entry.find("[name='your']").text(data.yourCallsign);

        entry.find("[name='myShort']").text(data.myCallsignSuffix);
        entry.find("[name='message']").text(data.message);
        entry.find("[name='lat']").text(data.latitude >= 0 ? data.latitude : "");
        entry.find("[name='lon']").text(data.longitude >= 0 ? data.longitude : "");

        $(entry).prependTo($(entry).parent());
    }
    else {
        const regResult = data.myCallsign.match(/^([A-Z0-9]{4,7})/g);
        const myCallsignNoModule = regResult != null && regResult.length >= 1 && regResult[0] ? regResult[0] : null;
        entry =
            $('<tr />').loadTemplate($("#template_jdstarnow_heard_entry"),{
                "index":0,
                "time":moment.unix(data.time / 1000).format("MM/DD HH:mm:ss"),
                "rpt1":data.repeater1Callsign,
                "rpt2":data.repeater2Callsign,
                "your":data.yourCallsign,
                "myLong":data.myCallsign,
                "myLongLink":myCallsignNoModule ? 'https://www.qrz.com/db/' + myCallsignNoModule : "#",
                "myShort":data.myCallsignSuffix,
                "message":data.message,
                "lat":data.latitude >= 0 ? data.latitude : "",
                "lon":data.longitude >= 0 ? data.longitude : ""
            });
        heardlog.prepend(entry);
    }

    const myLongCallsignLink = $(entry).find('td[name="myLong"]>a');
    if(data.transmitting){
        $(entry).addClass('table-success');
        $(entry).addClass('text-dark');
        $(entry).addClass('font-weight-bold');
        if(myLongCallsignLink.exists()){
            myLongCallsignLink.removeClass('text-light');
            myLongCallsignLink.addClass('text-danger');
        }
    }
    else{
        $(entry).removeClass('table-success');
        $(entry).removeClass('text-dark');
        $(entry).removeClass('font-weight-bold');
        if(myLongCallsignLink.exists()){
            myLongCallsignLink.addClass('text-light');
            myLongCallsignLink.removeClass('text-danger');
        }
    }

    const indexEntries = $(heardlog).find("th[name='index']");
    indexEntries.each(function(index, entry){
        if(index < dstarnowLimit)
            $(entry).text(index + 1);
        else
            $(entry).parent().remove();
    });
}

function setupNoraAPI(serverAddress, context) {
    noraapi = new NoraAPI(serverAddress, context, logLevel);

    noraapi.connect(
        () =>{
            noraapi.onNotifyLogEvent = onReceiveNotifyLog;

            setTimeout(() =>{
                noraapi.getDashboardInfo().then(onReceiveDashboardInfo);
    
                noraapi.enableDashboard('home', true);
    
                noraapi.addEventHandler(
                    'update_status_information.home',
                    onReceiveStatusInformation
                );

                clearTopMessage();

                showTopMessage("success", "Connected to server " + serverAddress + ".");
    
                setTimeout(() =>{
                    clearTopMessage();
                }, 10000);
            }, 1000);

            appendLogConsole("info", 'Client has connected to the server!');
        },
        () => {
            clearTopMessage();

            appendLogConsole('info', 'Client has disconnected, reloading...');

            setTimeout(() =>{
                location.reload();
            }, 5000);
        },
        () => {
            showTopMessage(
                "error",
                'Server ' + serverAddress + ' Connection timeout'
            );

            setTimeout(() =>{
                location.reload();
            }, 30000);
        },
        (err) => {
            showTopMessage(
                "error",
                'Server ' + serverAddress + ' Connection error = ' + err
            );

            setTimeout(() =>{
                location.reload();
            }, 30000);
        }
    );
}

function onReceiveStatusInformation(status) {
    logger.trace(JSON.stringify(status));

    const applicationInfo = $('#' + selectorEscape('home.application_info'));
    if(applicationInfo.exists()){
        applicationInfo.html(
            '<h5>' +
            htmlEntities(status.applicationName) + '@' +
            htmlEntities(status.applicationRunningOS) + '<br />v' +
            htmlEntities(status.applicationVersion) + '<br /><br />' +
            'on ' +
            htmlEntities(status.cpuUsageReport.operatingSystemName) + '<br />' +
            htmlEntities(status.cpuUsageReport.operatingSystemVersion) + '(' +
            htmlEntities(status.cpuUsageReport.arch) + ')' +
            '</h5>'
        );
    }

    const applicationUptime = $('#' + selectorEscape('home.application_uptime'));
    if(applicationUptime.exists()){

        const uptimeSeconds = Number(status.applicationUptime);
        const uptimeDays = Math.floor(uptimeSeconds / 86400);
        const uptimeHours = Math.floor((uptimeSeconds % 86400)  / 3600);
        const uptimeMinutes = Math.floor((uptimeSeconds % 86400 % 3600) / 60);
        applicationUptime.html(
            uptimeDays + 'd ' + uptimeHours + 'h ' + uptimeMinutes + 'm'
        );
    }

    const cpuLoadApp = $('#' + selectorEscape('home.cpu_load.app'));
    if(cpuLoadApp.exists()){
        let appLoadCurrent = 0.0;
        $.each(status.cpuUsageReport.threadUsageReport, (index, threadUsage) =>{
            appLoadCurrent += Number(threadUsage.cpuUsageCurrent);
        });
        appLoadCurrent = Math.floor(appLoadCurrent * 100);

        cpuLoadApp.text(appLoadCurrent + '%');
    }

    const cpuLoadSys = $('#' + selectorEscape('home.cpu_load.sys'));
    if(cpuLoadSys.exists()){
        const systemLoadAverage =
        Math.floor(Number(status.cpuUsageReport.systemLoadAverage) * 100);

        cpuLoadSys.text(systemLoadAverage + '%');
    }

    const gatewayOverview = $('#' + selectorEscape('home.gateway_overview'));
    if(gatewayOverview.exists()){
        gatewayOverview.empty();

        const gatewayStatus = status.gatewayStatusReport;
        const isScopePrivate = String(gatewayStatus.scope).toLowerCase() === 'private';
        const isScopePublic = String(gatewayStatus.scope).toLowerCase() === 'public';
        const isScopUnknown = String(gatewayStatus.scope).toLowerCase() === 'unknown';
        const isUseProxy = String(gatewayStatus.useProxy).toLowerCase() === 'true';

        const gateway = $('<tr />');
        gatewayOverview.append(gateway);
        gateway.append([
            $('<td />',{
                "html" : '<b>' + htmlEntities(gatewayStatus.gatewayCallsign) + '</b>'
            }),
            $('<td />', {
                "html" :
                    (
                        isUseProxy ?
                            htmlEntities(gatewayStatus.proxyServerAddress) + ' : ' +
                            htmlEntities(gatewayStatus.proxyServerPort) : 'DISABLED'
                    ),
                "class" : (isUseProxy ? 'text-warning' : 'text-primary')
            }),
            $('<td />',{"html" : htmlEntities(gatewayStatus.lastHeardCallsign)}),
            $('<td />',{
                "html" : htmlEntities(gatewayStatus.scope),
                "class" :
                    (isScopePrivate ? 'text-success' : '') +
                    (isScopePublic ? 'text-warning' : '') +
                    (isScopUnknown ? 'text-danger' : '')
            }),
            $('<td />',{"html" : htmlEntities(gatewayStatus.description1)}),
            $('<td />',{"html" : htmlEntities(gatewayStatus.description2)}),
            $('<td />',{
                "html" :
                    '<a class="text-white" href="' + gatewayStatus.url + '">' +
                    htmlEntities(gatewayStatus.url) + '</b>'
            })
        ]);
    }

    const repeaterOverview = $('#' + selectorEscape('home.repeater_overview'));
    if(repeaterOverview.exists()){
        repeaterOverview.empty();

        $.each(status.repeaterStatusReports, (index, repeaterStatus) =>{
            const isReflectorLinked =
                String(repeaterStatus.linkedReflectorCallsign).match(/[^ ]+/g);

            const isScopePrivate = String(repeaterStatus.scope).toLowerCase() === 'private';
            const isScopePublic = String(repeaterStatus.scope).toLowerCase() === 'public';
            const isScopUnknown = String(repeaterStatus.scope).toLowerCase() === 'unknown';

            const repeater = $('<tr />');
            repeaterOverview.append(repeater);
            repeater.append([
                $('<td />',{
                    "name" : 'repeaterCallsign',
                    "html" : '<b>' + htmlEntities(repeaterStatus.repeaterCallsign) + '</b>'
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.repeaterType)
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.routingService)
                }),
                $('<td />' ,{
                    "html" :
                        (isReflectorLinked ? htmlEntities(repeaterStatus.linkedReflectorCallsign) : '-'),
                    "class" : (isReflectorLinked ? 'text-danger table-success font-weight-bold' : '')
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.lastHeardCallsign)
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.scope),
                    "class" :
                        (isScopePrivate ? 'text-success' : '') +
                        (isScopePublic ? 'text-warning' : '') +
                        (isScopUnknown ? 'text-danger' : '')
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.description1)
                }),
                $('<td />' ,{
                    "html" : htmlEntities(repeaterStatus.description2)
                }),
                $('<td />' ,{
                    "html" : '<a class="text-white" href="' + repeaterStatus.url + '">' + htmlEntities(repeaterStatus.url) + '</b>'
                })
            ]);
        });

        repeaterOverview.children().sort((a, b) =>{
            if($(a).find('td[name="repeaterCallsign"]').text() > $(b).find('td[name="repeaterCallsign"]').text())
                return 1;
            else
                return -1;
        }).appendTo(repeaterOverview);
    }

    const reflectorOverview = $('#' + selectorEscape('home.reflector_overview'));
    if(reflectorOverview.exists()){
        reflectorOverview.empty();

        $(status.reflectorStatusReports).each((index, reflectorStatus) =>{
            const isIncomingEnabled =
                String(reflectorStatus.enableIncomingLink).toLowerCase() === 'true';
            const isOutgoingEnabled =
                String(reflectorStatus.enableOutgoingLink).toLowerCase() === 'true';
            const reflector = $('<tr />');
            reflectorOverview.append(reflector);
            reflector.append([
                $('<td />' ,{
                    "name" : 'reflectorType',
                    "html" : '<b>' + htmlEntities(reflectorStatus.reflectorType) + '</b>'
                }),
                $('<td />' ,{
                    "html" : htmlEntities(reflectorStatus.serviceStatus)
                }),
                $('<td />' ,{
                    "html" :
                        (isIncomingEnabled ? "ENABLED" : "DISABLED") +
                        ' : ' + htmlEntities(reflectorStatus.incomingLinkPort),
                    "class" : (isIncomingEnabled ? 'text-warning' : 'text-success')
                }),
                $('<td />' ,{
                    "html" : htmlEntities(reflectorStatus.connectedIncomingLink)
                }),
                $('<td />' ,{
                    "html" : (isOutgoingEnabled ? "ENABLED" : "DISABLED"),
                    "class" : (isOutgoingEnabled ? 'text-warning' : 'text-success')
                }),
                $('<td />' ,{
                    "html" : htmlEntities(reflectorStatus.connectedOutgoingLink)
                })
            ]);
        });

        reflectorOverview.children().sort((a, b) =>{
            if($(a).find('[name="reflectorType"]').text() > $(b).find('[name="reflectorType"]').text())
                return 1;
            else
                return -1;
        }).appendTo(reflectorOverview);
    }

    const routingOverview = $('#' + selectorEscape('home.routing_overview'));
    if(routingOverview.exists()){
        routingOverview.empty();

        $(status.routingStatusReports).each((index, routingStatus) => {
            const routing = $('<tr />');
            routingOverview.append(routing);
            routing.append([
                $('<td />', {
                    "html" : '<b>' + htmlEntities(routingStatus.serviceType) + '</b>'
                }),
                $('<td />', {
                    "html" : htmlEntities(routingStatus.serviceStatus)
                }),
                $('<td />', {
                    "html" :
                        htmlEntities(routingStatus.serverAddress) + ' : ' +
                        htmlEntities(routingStatus.serverPort)
                })
            ]);
        });
    }
}

function onReceiveDashboardInfo(info) {
    dashboardInfo = info;

    createNavbar(info);
    createContents(info);

    updateView();
}

function onReceiveNotifyLog(data) {
    let level;
    if(data.level.includes("TRACE"))
        level = "trace";
    else if(data.level.includes("DEBUG"))
        level = "debug";
    else if(data.level.includes("INFO"))
        level = "info";
    else if(data.level.includes("WARN"))
        level = "warn";
    else if(data.level.includes("ERROR"))
        level = "error";
    else
        level = "info";

    appendLogConsole(level, data);
}

function createNavbar(data) {

    const gatewayInfo = data.gatewayInfo;
    const repeaterInfos = data.repeaterInfos;

    const gatewayCallsign =
        gatewayInfo.gatewayCallsign
        .substr(0, gatewayInfo.gatewayCallsign.length - 1).trim();
    const title = gatewayCallsign + " Dashboard";

    $('title').text(title);

    $('#main_navbar_brand').text(title);
    const navbarText = $('#main_navbar_text');
    if(navbarText.exists()){
        navbarText.html(
            data.applicationName +
            " v" + data.applicationVersion +
            "@" + data.applicationRunningOS +
            ' / ' +
            navbarText.html()
        );
    }


    $('#main_navbar_item_gateway_link').attr('href', "#" + gatewayInfo.webSocketRoomId);

    repeaterInfos.forEach(repeaterInfo => {
        const repeaterItem = $("<li/>",{"class":"nav-item dropdown"});
        repeaterItem.append(
            $("<a/>",{
                "class":"dropdown-item dropdown-toggle",
                "data-toggle":"dropdown",
                "aria-haspopup":true,
                "aria-expanded":false,
                href : "#" + repeaterInfo.webSocketRoomId,
                text : repeaterInfo.repeaterCallsign
            })
        );
        
        const modemItems =
            $('<ul/>', {
                "class":"dropdown-menu"
            });
        
        modemItems.append(
            $('<li/>').append(
                $('<a/>', {
                        "class":"dropdown-item",
                        "data-toggle" : "tab",
                        href:"#" + repeaterInfo.webSocketRoomId,
                        text:repeaterInfo.repeaterCallsign
                    })
            )	
        );

        modemItems.append(
            $('<div/>', {"class":"dropdown-divider"})
        );

        repeaterInfo.modemInfos.forEach(modemInfo => {
            modemItems.append(
                $('<li/>').append(
                    $('<a/>', {
                        "class":"dropdown-item",
                        href:"#" + modemInfo.webSocketRoomId,
                        text:modemInfo.modemType + "(" + modemInfo.modemId + ")"
                    })
                )
            );

        });
        
        repeaterItem.append(modemItems);
        
        $('#main_navbar_item_repeaters_ul').append(repeaterItem);
    });

    $('#main_navbar').bootnavbar();
}

function createContents(data) {
    const maintab = $('#maintab');

    const gatewayTab =
        $('<div/>',{"id":"maintab_" + data.gatewayInfo.webSocketRoomId,"class":"tab-pane"});
    maintab.append(gatewayTab);

    noraapi.addEventHandler(
        "notify_updateheard_" + data.gatewayInfo.webSocketRoomId,
        (data) => {
            const gatewayTabID =
                "#maintab_" + selectorEscape(dashboardInfo.gatewayInfo.webSocketRoomId);
            const gatewayTab = $(gatewayTabID);
            if(gatewayTab.exists())
                updateContentsGatewayHeardEntry(gatewayTab, data, true);
            else
                logger.error("Gateway tab not found = " + gatewayTabID + ".");
        }
    );

    noraapi.addEventHandler(
        "notify_update_reflectorhosts_" + data.gatewayInfo.webSocketRoomId,
        (data) => {
            const gatewayTabID =
                "#maintab_" + selectorEscape(dashboardInfo.gatewayInfo.webSocketRoomId);
            const gatewayTab = $(gatewayTabID);
            if(gatewayTab.exists())
                updateContentsGatewayReflectorHosts(gatewayTab, data, true);
            else
                logger.error("Gateway tab not found = " + gatewayTabID + ".");
        }
    );

    data.repeaterInfos.forEach(repeaterInfo => {
        maintab.append(
            $('<div/>',{"id":"maintab_" + repeaterInfo.webSocketRoomId,"class":"tab-pane"})
            .append($('<p/>',{text:repeaterInfo.webSocketRoomId}))
        );

        repeaterInfo.modemInfos.forEach(modemInfo =>{
            if(modemInfo.webSocketRoomId){
                const modemtab = 
                    $('<div/>',{
                        "id":"maintab_" + modemInfo.webSocketRoomId,
                        "class":"tab-pane"
                    });

                maintab.append(modemtab);

                if(modemInfo.modemType === 'AnalogModemPiGPIO'){

                    noraapi.addEventHandler("update_status_" + modemInfo.webSocketRoomId,(data) =>{
                        const maintabId = "#maintab_" + selectorEscape(data.webSocketRoomId);
                        const maintab = $(maintabId);
                        if(maintab.exists())
                            updateContentsAnalogModemPiGPIO(maintab, data);
                        else
                            logger.error("Modem tab not found = " + maintabId + ".");
                    });

                    updateContentsAnalogModemPiGPIO(modemtab,{
                        "webSocketRoomId":modemInfo.webSocketRoomId,
                        "gatewayCallsign":data.gatewayInfo.gatewayCallsign,
                        "repeaterCallsign":repeaterInfo.repeaterCallsign,
                        "uplinkActive":false,
                        "downlinkActive":false,
                        "uplinkHeader":{
                            "flags":[0,0,0],
                            "repeater1Callsign":"",
                            "repeater2Callsign":"",
                            "yourCallsign":"",
                            "myCallsignLong":"",
                            "myCallsignShort":""
                        },
                        "downlinkHeader":{
                            "flags":[0,0,0],
                            "repeater1Callsign":"",
                            "repeater2Callsign":"",
                            "yourCallsign":"",
                            "myCallsignLong":"",
                            "myCallsignShort":""
                        },
                        "uplinkConfigUseGateway":false,
                        "uplinkConfigYourCallsign":"",
                        "uplinkConfigMyCallsign":"",
                    });
                }
            }

            //maintab.append(modemtab);
        });
    });
}

function getContentsGateway(gatewayTab, functionID) {
    let gatewayContents =
        $(gatewayTab).find(
            "[contents-id=" + selectorEscape(functionID) + "]"
        );
    if(!gatewayContents.exists()){
        gatewayContents =
            $('<div />').loadTemplate($("#template_Gateway"),{
                "contents_id" : functionID,
                "lastheard_id" : functionID + '.lastheard',
                "reflectorhosts_id" : functionID + '.reflectorhosts'
            });
        gatewayTab.append(gatewayContents);
    }

    return gatewayContents;
}

function updateContentsGatewayReflectorHosts(gatewayTab, data, notifyUpdate) {
//    logger.info(data);

    const functionID = dashboardInfo.gatewayInfo.webSocketRoomId;

    const gatewayContents = getContentsGateway(gatewayTab, functionID);

    const reflectorHosts =
        gatewayContents.find('#' + selectorEscape(functionID + '.reflectorhosts'));
    if(!reflectorHosts.exists()){
        logger.error("ReflectorHosts not found.");
        return;
    }

    if(!notifyUpdate)
        reflectorHosts.empty();

    const newHosts = [];

    $(data).each((index, host) =>{
        let entry =
            $(reflectorHosts).find("td[name='callsign']")
            .filter(function() {
                return $(this).text() === host.reflectorCallsign;
            }).parent();

        if(entry.exists()){
            entry.find('[name="protocol"]').text(host.reflectorProtocol);
            entry.find('[name="address"]').text(
                host.reflectorAddress + ' : ' + host.reflectorPort
            );
            entry.find('[name="priority"]').text(host.priority);
            entry.find('[name="updateTime"]').text(
                moment.unix(host.updateTime / 1000).format("MM/DD HH:mm:ss")
            );
        }
        else {
            entry =
                $('<tr />', {"class" : "text-white"}).append([
                    $('<th />', {
                        "name" : "index",
                        "scope" : "row",
                        "html" : 0
                    }),
                    $('<td />', {
                        "name" : "callsign",
                        "html" : host.reflectorCallsign
                    }),
                    $('<td />', {
                        "name" : "protocol",
                        "html" : host.reflectorProtocol
                    }),
                    $('<td />', {
                        "name" : "address",
                        "html" : host.reflectorAddress + ' : ' + host.reflectorPort
                    }),
                    $('<td />', {
                        "name" : "priority",
                        "html" : host.priority
                    }),
                    $('<td />', {
                        "name" : "updateTime",
                        "html" : moment.unix(host.updateTime / 1000).format("MM/DD HH:mm:ss")
                    })
                ]);

            newHosts.push(entry);
        }
    });

    if(newHosts.length >= 1)
        $(reflectorHosts).append(newHosts);

    $(reflectorHosts).children().sort((a, b) =>{
        if($(a).find('[name="callsign"]').text() > $(b).find('[name="callsign"]').text())
            return 1;
        else
            return -1;
    }).appendTo(reflectorHosts);

    $(reflectorHosts).children().each((index, entry) =>{
        $(entry).find('[name="index"]').text(index + 1);
    });
}

function updateContentsGatewayHeardEntry(gatewayTab, data, update) {
    const functionID = dashboardInfo.gatewayInfo.webSocketRoomId;

    const gatewayContents = getContentsGateway(gatewayTab, functionID);

    const heardlog =
        gatewayContents.find('#' + selectorEscape(functionID + '.lastheard'));
    if(!heardlog.exists()){
        logger.error("HeardLog not found.");
        return;
    }

    let entry =
        $(heardlog).find("td[name='myLong']")
        .find("a")
        .filter(function() {
            return $(this).text() === data.myCallsignLong;
        }).parent();
    if(entry.exists()){
        entry = entry.parent();
        
        entry.find("[name='time']").text(
            moment.unix(data.heardTime).format("MM/DD HH:mm:ss")
        );
        entry.find("[name='dir']").text(data.direction);
        entry.find("[name='rpt1']").text(data.repeater1Callsign);
        entry.find("[name='rpt2']").text(data.repeater2Callsign);
        entry.find("[name='your']").text(data.yourCallsign);

        entry.find("[name='myShort']").text(data.myCallsignShort);
        entry.find("[name='from']").text(data.from);
        entry.find("[name='dest']").text(data.destination);
        entry.find("[name='message']").text(data.shortMessage);
        entry.find("[name='lat']").text(data.locationAvailable ? data.latitude : "");
        entry.find("[name='lon']").text(data.locationAvailable ? data.longitude : "");
        entry.find("[name='proto']").text(data.protocol);

        $(entry).prependTo($(entry).parent());
    }
    else if(!update || (update && data.state === "Start")) {
        const regResult = data.myCallsignLong.match(/^([A-Z0-9]{4,7})/g);
        const myCallsignNoModule = regResult[0] ? regResult[0] : null;
        entry =
            $(
                '<tr />',
                {
                    "class" : "text-light"
                }
            ).loadTemplate($("#template_gateway_lastheard_entry"),{
                "index":0,
                "time":moment.unix(data.heardTime).format("MM/DD HH:mm:ss"),
                "dir":data.direction,
                "rpt1":data.repeater1Callsign,
                "rpt2":data.repeater2Callsign,
                "your":data.yourCallsign,
                "myLong":data.myCallsignLong,
                "myLongLink":myCallsignNoModule ? 'https://www.qrz.com/db/' + myCallsignNoModule : "#",
                "myShort":data.myCallsignShort,
                "from":data.from,
                "dest":data.destination,
                "message":data.shortMessage,
                "lat":data.locationAvailable ? data.latitude : "",
                "lon":data.locationAvailable ? data.longitude : "",
                "proto":data.protocol
            });
        heardlog.prepend(entry);
    }

    const myLongCallsignLink = $(entry).find('td[name="myLong"]>a');
    if(data.state === "Start" || data.state === "Update"){
        if(data.direction === "OUTGOING")
            $(entry).addClass('table-danger');
        else if(data.direction === "INCOMING")
            $(entry).addClass('table-success');

        $(entry).removeClass("text-light");
        $(entry).addClass("text-dark");
        $(entry).addClass("font-weight-bold");
        if(myLongCallsignLink.exists()){
            $(myLongCallsignLink).removeClass("text-light");
            $(myLongCallsignLink).addClass("text-danger");
        }
    }
    else{
        $(entry).removeClass('table-danger');
        $(entry).removeClass('table-success');

        $(entry).removeClass("text-dark");
        $(entry).addClass("text-light");
        $(entry).removeClass("font-weight-bold");

        if(myLongCallsignLink.exists()){
            $(myLongCallsignLink).removeClass("text-danger");
            $(myLongCallsignLink).addClass("text-light");
        }
    }

    const entries = $(heardlog).find("th[name='index']");
    entries.each(function(index, entry){
        if(index < lastHeardLimit)
            $(entry).text(index + 1);
        else
            $(entry).parent().remove();
    });
}

function updateContentsGatewayHeardLog(gatewayTab, data) {
    logger.trace(
        "Updating contents gateway heardlog.\n" +
        JSON.stringify(data.heardLog)
    );

    data.heardLog.forEach(function(entry, index){
        updateContentsGatewayHeardEntry(gatewayTab, entry, false)
    });
}

function updateContentsAnalogModemPiGPIO(modemtab, data) {
    logger.trace(
        "Update contents AnalogModemPiGPIO.\n" +
        JSON.stringify(data)
    );

    modemtab.empty();

    modemtab.append(
        $('<div/>').loadTemplate($("#template_AnalogModemPiGPIO"),{
            "contents_id":data.webSocketRoomId,
            "downlink_yourcallsign":data.downlinkActive ? data.downlinkHeader.yourCallsign : "",
            "downlink_mycallsign":data.downlinkActive ? data.downlinkHeader.myCallsignLong + "/" + data.downlinkHeader.myCallsignShort : "",
            "uplink_usegateway":undefined,
            "uplink_yourcallsign":data.uplinkActive ? data.uplinkHeader.yourCallsign : "",
            "uplink_mycallsign":data.uplinkActive ? data.uplinkHeader.myCallsignLong + "/" + data.uplinkHeader.myCallsignShort : "",
            "uplink_config_usegateway":data.uplinkConfigUseGateway,
            "uplink_config_yourcallsign":data.uplinkConfigYourCallsign.trim(),
            "uplink_config_mycallsign":data.uplinkConfigMyCallsign.trim(),
            "uplink_config_form_id":data.webSocketRoomId + ".uplink_config_form",
            "uplink_config_usegateway_id":data.webSocketRoomId + ".uplink_config_usegateway_checkbox",
            "uplink_config_yourcallsign_id":data.webSocketRoomId + ".uplink_config_youcallsign_input",
            "uplink_config_mycallsign_id":data.webSocketRoomId + ".uplink_config_mycallsign_input",
            "uplink_config_update_button_id":data.webSocketRoomId + ".uplink_config_update_button"
        })
    );

    const updateConfigUpdateButton =
        $("#" + selectorEscape(data.webSocketRoomId) + "\\.uplink_config_update_button");
    if(!updateConfigUpdateButton.exists()){
        logger.error("Could not found updateConfigUpdateButton");

        return;
    }

    updateConfigUpdateButton.on('click', function() {
        logger.trace("Update button click() " + $(this));

        const contentsID = $(this).parents('[contents-id]');
        logger.trace("ContentsID = " + contentsID);
        if(!contentsID.exists()){
            logger.error("Could not fould contents-id");

            return;
        }
        const webSocketRoomId = contentsID.attr('contents-id');
        const info = findGatewayRepeaterInfoByModemWebSocketRoomID(webSocketRoomId);
        if(!info){return;}

        const useGateway =
            $("#" + selectorEscape(webSocketRoomId)+ "\\.uplink_config_usegateway_checkbox").prop("checked");
        const yourCallsign =
            $("#" + selectorEscape(webSocketRoomId) + "\\.uplink_config_youcallsign_input").val();
        const myCallsign =
            $("#" + selectorEscape(webSocketRoomId) + "\\.uplink_config_mycallsign_input").val();

        noraapi.updateUplinkHeader(
            "update_uplink_header_" + webSocketRoomId,
            {
                "flags":[0,0,0],
                "repeater1Callsign":info.repeaterInfo.repeaterCallsign,
                "repeater2Callsign":useGateway?info.gatewayInfo.gatewayCallsign:info.repeaterInfo.repeaterCallsign,
                "yourCallsign":yourCallsign,
                "myCallsignLong":myCallsign,
                "myCallsignShort":"NVRA"
            }
        );
    });
}

function findGatewayRepeaterInfoByModemWebSocketRoomID(webSocketRoomId) {
    const gatewayInfo = dashboardInfo.gatewayInfo;
    let repeaterInfo, modemInfo;
    repeaterInfo = dashboardInfo.repeaterInfos.find(repeaterInfo => {
        modemInfo = repeaterInfo.modemInfos.find(modemInfo => {
            return modemInfo.webSocketRoomId === webSocketRoomId;
        });
        return modemInfo;
    });

    if(gatewayInfo && repeaterInfo && modemInfo)
        return {'gatewayInfo':gatewayInfo,'repeaterInfo':repeaterInfo,'modemInfo':modemInfo};
    else{
        logger.error("Could not found websocketroomid = " + webSocketRoomId);

        return null;
    }
}

function updateView() {
    if(currentSelectedFunction === 'home')
        $('#main_navbar_item_home').addClass('active');
    else
        $('#main_navbar_item_home').removeClass('active');

    if(currentSelectedFunction === dashboardInfo.gatewayInfo.webSocketRoomId)
        $('#main_navbar_item_gateway').addClass("active");
    else
        $('#main_navbar_item_gateway').removeClass("active");

    if(dashboardInfo.repeaterInfos.find((repeaterInfo) => {
        const result =
            currentSelectedFunction === repeaterInfo.webSocketRoomId ||
            repeaterInfo.modemInfos.find((modemInfo) => {
                return currentSelectedFunction === modemInfo.webSocketRoomId
            });

        return result;
    }))
        $('#main_navbar_item_repeaters').addClass("active");
    else
        $('#main_navbar_item_repeaters').removeClass("active");		

    const breadcrumb = $('#breadcrumb');
    breadcrumb.empty();
    breadcrumb.append(
        $('<li/>',{
            "class":"breadcrumb-item"
        }).append($('<a/>',{
            "class" : "text-white font-weight-bold",
            href:"#home",
            text:"HOME >"
        }))
    );
    if(currentSelectedFunction.match(dashboardInfo.gatewayInfo.webSocketRoomId)){
        breadcrumb.append(
            $('<li/>',{
                "class":"breadcrumb-item"
            }).append($('<a/>',{
                "class" : "text-white font-weight-bold",
                href:"#" + dashboardInfo.gatewayInfo.webSocketRoomId,
                text:dashboardInfo.gatewayInfo.gatewayCallsign + " >"
            }))
        );
    }
    dashboardInfo.repeaterInfos.forEach((repeaterInfo) => {
        if(currentSelectedFunction.match(repeaterInfo.webSocketRoomId)){
            breadcrumb.append(
                $('<li/>',{
                    "class":"breadcrumb-item"
                }).append($('<a/>',{
                    "class" : "text-white font-weight-bold",
                    href:"#" + repeaterInfo.webSocketRoomId,
                    text:repeaterInfo.repeaterCallsign + " >"
                }))
            );
        }

        const modem =
            repeaterInfo.modemInfos.find((modemInfo) =>{
                return currentSelectedFunction === modemInfo.webSocketRoomId;
            });
        if(modem) {
            breadcrumb.append(
                $('<li/>',{
                    "class":"breadcrumb-item"
                }).append($('<a/>',{
                    "class" : "text-white font-weight-bold",
                    href:"#" + modem.webSocketRoomId,
                    text:modem.modemType + "(" + modem.modemId + ")" + " >"
                }))
            );
            return false;
        }
    });

    $('#maintab div[class*="tab-pane"').each((index, tabPanel) =>{
        if(tabPanel.id === ("maintab_" + currentSelectedFunction))
            $(tabPanel).addClass('active');
        else
            $(tabPanel).removeClass('active');
    });
}

$(document).on('click', '.navbar a, .breadcrumb a', function(){
    const selectedFunction = $(this).attr("href").replace(/[^a-z0-9_.]/g, '');
    logger.trace("Navbar clicked = " + selectedFunction);

    if(selectedFunction){
        const roomChanged = currentSelectedFunction !== selectedFunction;
        if(roomChanged)
            onFunctionChange(selectedFunction, currentSelectedFunction);
    }
});

function onFunctionChange(currentFunc, prevFunc) {
    noraapi.enableDashboard(prevFunc, false);
    noraapi.enableDashboard(currentFunc, true);

    logger.debug("Change room to " + currentFunc + '<-' + prevFunc);

    currentSelectedFunction = currentFunc;

    updateView();

    noraapi.requestStatus(currentFunc);

    if(dashboardInfo.gatewayInfo.webSocketRoomId === currentFunc){
        const funcID = dashboardInfo.gatewayInfo.webSocketRoomId;
        
        const gatewayTabID =
            "#maintab_" + selectorEscape(dashboardInfo.gatewayInfo.webSocketRoomId);
        const gatewayTab = $(gatewayTabID);
        if(!gatewayTab.exists())
            logger.error("Gateway tab not found = " + gatewayTabID + ".");
        
        noraapi.getHeardLog(funcID).then((data) => {
            updateContentsGatewayHeardLog(gatewayTab, data);
        });

        noraapi.getReflectorHosts(funcID).then((data) => {
            updateContentsGatewayReflectorHosts(gatewayTab, data, false)
        });
    }
}

function appendLogConsole(category, message) {
    const logConsole = $('#log_console');
    if(!logConsole.exists()){return;}

    while(logConsole.children().length >= 100)
        logConsole.children().last().remove();

    let logType;
    switch(category){
        case "trace":
            logType = "log-trace-msg";
            break;
        case "debug":
            logType = "log-debug-msg";
            break;
        case "info":
            logType = "log-info-msg";
            break;
        case "warn":
            logType = "log-warn-msg";
            break;
        case "error":
            logType = "log-error-msg";
            break;
        default:
            logType = "log-info-msg";
            break;
    }

    const element =
        $('<div />').addClass('d-table-row').loadTemplate($("#template_LogEntry"), {
            "levelClass":logType,
            "level":message.level ? message.level : category.toUpperCase(),
            "time":moment.unix(message.timestamp ? message.timestamp : moment().unix()).format("YYYY/MM/DD HH:mm:ss"),
            "message":message.message ? message.message.replace(/\r?\n/g, '<br>') : message
        });

    logConsole.prepend(element);
}
